#=========================================================================
# Copyright (C) 2017 Intel Corporation
#
# Licensed under the Apache License,  Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law  or agreed  to  in  writing,  software
# distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the  specific  language  governing  permissions  and
# limitations under the License.
#=========================================================================

#
# Intel(R) Cryptography Primitives Library
#

import re

def readNextFunction(header, curLine, headerID):    ## read next function with arguments
  ## find header ID macros
  FunName  = ''
  FunArg   = ''
  FunType  = ''
  success = False
  while (curLine < len(header) and success == False):
    if not headerID and re.match(r'\s*#\s*if\s*!\s*defined\s*\(\s*__IPP', header[curLine]):
      headerID= re.sub(r'.*__IPP', '__IPP', header[curLine] )
      headerID= re.sub(r'\)', '', headerID)
      headerID= re.sub(r'[\n\s]', '', headerID )

    if re.match(r'^\s*IPPAPI\s*\(.*', header[curLine] ) :
      FunStr= header[curLine];
      FunStr= re.sub(r'\n','',FunStr)   ## remove EOL symbols

      while not re.match(r'.*\)\s*\)\s*$', FunStr):   ## concatenate string if string is not completed
        curLine= curLine+1
        FunStr= FunStr+header[curLine]
        FunStr= re.sub(r'\n','',FunStr)   ## remove EOL symbols

      FunStr= re.sub(r'\s+', ' ', FunStr)

      s= FunStr.split(',')

      ## Extract function name
      FunName= s[1]
      FunName= re.sub(r'\s', '', FunName)

      ## Extract function type
      FunType= re.sub(r'.*\(', '', s[0] )
      #FunType= re.sub(r' ', '', FunType )

      ## Extract function arguments
      FunArg= re.sub(r'.*\(.*,.+,\s*\(', '(', FunStr)
      FunArg= re.sub(r'\)\s*\)', ')', FunArg)
      success = True

    curLine = curLine + 1

  return {'curLine':curLine, 'FunType':FunType, 'FunName':FunName, 'FunArg':FunArg, 'success':success }



"""
Copyright (C) 2019 Intel Corporation

SPDX-License-Identifier: MIT
"""

import os
import shutil
import unittest
from sys import platform

import tests.utils
from tool.core import build, generate_script


class FunctionalityTests(unittest.TestCase):
    functions = ["ippccGetLibVersion", "ippdcGetLibVersion", "ippsGetLibVersion"]
    domains = ["ippcc", "ippdc", "ipps", "ippcore", "ippvm"]

    def generator_assertion(self):
        self.assertTrue(os.path.exists("./tmp"), "Temporary folder does not exist")
        self.assertTrue(os.path.exists("./tmp/main.c"), "main.c file was not generated")
        self.assertTrue(
            os.path.exists("./tmp/" + tests.utils.EXPORT_FILES[tests.utils.HOST_SYSTEM]),
            "Export file was not generated",
        )
        self.assertTrue(
            os.path.exists("./tmp/" + tests.utils.BUILD_SCRIPT[tests.utils.INTEL64][tests.utils.HOST_SYSTEM]),
            "Build script was not generated",
        )

    @classmethod
    def setUpClass(cls):
        if platform == "linux" or platform == "linux2":
            tests.utils.HOST_SYSTEM = tests.utils.LINUX
        elif platform == "win32":
            tests.utils.HOST_SYSTEM = tests.utils.WINDOWS

    @classmethod
    def tearDown(self):
        if os.path.exists(tests.utils.TEMPORARY_FOLDER):
            shutil.rmtree(tests.utils.TEMPORARY_FOLDER, ignore_errors=True)

    def test_generation(self):
        generate_script(
            tests.utils.HOST_SYSTEM,
            tests.utils.HOST_SYSTEM,
            self.functions,
            tests.utils.TEMPORARY_FOLDER,
            "tmp_dll",
            self.domains,
            tests.utils.INTEL64,
            False,
        )
        self.generator_assertion()

    def test_build(self):
        self.assertTrue(
            build(
                tests.utils.HOST_SYSTEM,
                tests.utils.HOST_SYSTEM,
                self.functions,
                os.path.abspath(tests.utils.TEMPORARY_FOLDER),
                "tmp_dll",
                self.domains,
                tests.utils.INTEL64,
                False,
                os.environ["COMPILERS_AND_LIBRARIES"],
            )
        )


if __name__ == "__main__":
    unittest.main()



"""
Copyright (C) 2019 Intel Corporation

SPDX-License-Identifier: MIT
"""

HOST_SYSTEM = ""

WINDOWS = "Windows"
LINUX = "Linux"

TEMPORARY_FOLDER = "./tmp"

INTEL64 = "intel64"

BUILD_SCRIPT = {
    INTEL64: {WINDOWS: "intel64.bat", LINUX: "intel64.sh"},
}

EXPORT_FILES = {WINDOWS: "export.def", LINUX: "export.def"}

LIBRARIES_EXTENSIONS = {WINDOWS: ".dll", LINUX: ".so"}

LIBRARIES_PREFIX = {WINDOWS: "", LINUX: "lib"}



"""
Copyright (C) 2018 Intel Corporation

SPDX-License-Identifier: MIT
"""

from tool.utils import INTEL64, IPP, IPPCP, LINUX, OPENMP, SINGLE_THREADED, TBB, WINDOWS

ENV_VAR = {WINDOWS: "%{env_var}%", LINUX: "${{{env_var}}}"}

CALL_ENV_SCRIPT_COMMAND = {
    WINDOWS: 'call "{env_script}" {arch} {force_flag}',
    LINUX: 'source "{env_script}" -arch {arch} {force_flag}',
}

SET_ENV_COMMAND = {
    WINDOWS: 'set "{env_var}={path}"',
    LINUX: 'export "{env_var}={path}"',
}

ADDITIONAL_ENV = {
    WINDOWS: "",
    LINUX: "export LIBRARY_PATH=$LD_LIBRARY_PATH:$LIBRARY_PATH",
}

COMPILERS = {WINDOWS: "cl.exe", LINUX: "g++"}

LINKERS = {WINDOWS: "link.exe", LINUX: "g++"}

COMPILERS_FLAGS = {
    WINDOWS: {INTEL64: "/c /MP /MT /GS /sdl /O2"},
    LINUX: {
        INTEL64: (
            "-c -m64 -fPIC -fPIE -fstack-protector-strong -fstack-protector -O2 -D_FORTIFY_SOURCE=2 "
            "-fcf-protection -Wformat -Wformat-security"
        ),
    },
}

LINKER_FLAGS = {
    WINDOWS: {
        INTEL64: "/MACHINE:X64 /NXCompat /DynamicBase /CETCOMPAT /DEPENDENTLOADFLAG:0x2000",
    },
    LINUX: {
        INTEL64: "-z noexecstack -z relro -z now -fcf-protection",
    },
}

COMPILE_COMMAND_FORMAT = {
    WINDOWS: "{compiler} {cmp_flags} " '/I "%{root_type}%\\include" ' "{c_files}",
    LINUX: "{compiler} {cmp_flags}" ' -I "${root_type}/include" ' "{c_files}",
}

LINK_COMMAND_FORMAT = {
    WINDOWS: "{linker} /DLL {link_flags} /VERBOSE:SAFESEH "
    '/DEF:"{export_file}" *.obj '
    '/OUT:"{custom_library}.dll" /IMPLIB:"{custom_library}.lib" '
    "{ipp_libraries} "
    "{exp_libs}",
    LINUX: "{linker} -shared {link_flags} "
    '"{export_file}" *.o '
    '-o "{custom_library}.so" '
    "{ipp_libraries} "
    '-L"{sys_libs_path}" -lc -lm {exp_libs}',
}

SYS_LIBS_PATH = {
    WINDOWS: {
        INTEL64: "",
    },
    LINUX: {INTEL64: "$SYSROOT/lib64"},
}

EXP_LIBS = {
    WINDOWS: {SINGLE_THREADED: "", TBB: "tbb.lib", OPENMP: "libiomp5md.lib"},
    LINUX: {SINGLE_THREADED: "", TBB: "-ltbb", OPENMP: "-liomp5"},
}

MAIN_FILE = {
    WINDOWS: r"""#include <Windows.h>
#include "{package_type}.h"

int WINAPI DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{{
    switch (fdwReason)
    {{
    case DLL_PROCESS_ATTACH:
        {package_type}Init(); break;
    case DLL_THREAD_ATTACH: break;
    case DLL_THREAD_DETACH: break;
    case DLL_PROCESS_DETACH: break;
    default: break;
    }}
    return 1;
    UNREFERENCED_PARAMETER(hinstDLL);
    UNREFERENCED_PARAMETER(lpvReserved);
}}
""",
    LINUX: r"""#include "{package_type}.h"

__attribute__((constructor))
void init_custom_library(void)
{{
    (void){package_type}Init();
}}
""",
}

CUSTOM_DISPATCHER_FILE = r"""{include_lines}

#ifndef IPP_CALL
#   define IPP_CALL IPP_STDCALL
#endif

#define IPPFUN(type, name, arg) extern type IPP_CALL name arg

#ifndef NULL
#   ifdef __cplusplus
#       define NULL 0
#   else
#       define NULL ((void*)0)
#   endif
#endif

{architecture}

{features}{uninitialized_features_define}

#ifdef __cplusplus
extern "C" {{
#endif

{dispatcher}

#ifdef __cplusplus
}}
#endif

#endif
"""

RENAME_FORMAT = r"""
{declaration}
#define {function} {prefix}{function}
"""

RENAME_HEADER_SCHEME = r"""#ifndef {prefix}_RENAME_H
#define {prefix}_RENAME_H

{defines_header_statement}

#ifdef __cplusplus
extern "C" {{
#endif
{rename_statements}
#ifdef __cplusplus
}}
#endif

#endif // {prefix}_RENAME_H
"""

INCLUDE_STR = "#include <{header_name}>"

ARCHITECTURE_DEFINE = {
    INTEL64: "#if defined (_M_AMD64) || defined (__x86_64__)",
}

FEATURES_AVX3_COMMON_INTEL64 = r"""
#define AVX3I_FEATURES (ippCPUID_SHA | ippCPUID_AVX512VBMI | ippCPUID_AVX512VBMI2 | ippCPUID_AVX512IFMA | \
    ippCPUID_AVX512GFNI | ippCPUID_AVX512VAES | ippCPUID_AVX512VCLMUL)
#define AVX3X_FEATURES (ippCPUID_AVX512F | ippCPUID_AVX512CD | ippCPUID_AVX512VL | ippCPUID_AVX512BW | \
    ippCPUID_AVX512DQ)
"""

CPUID_FEATURES = {
    IPP: FEATURES_AVX3_COMMON_INTEL64 + r"""#define APX_FEATURES   (AVX3X_FEATURES | ippCPUID_AVX10_2)""",
    IPPCP: FEATURES_AVX3_COMMON_INTEL64 + r"""#define APX_FEATURES   (AVX3I_FEATURES | ippCPUID_AVX10_2)""",
}

UNINITIALIZED_FEATURES = {IPP: "(ippCPUID_MMX | ippCPUID_SSE | ippCPUID_SSE2)", IPPCP: "0"}

UNINITIALIZED_FEATURES_DEFINE = r"""
#define UNINITIALIZED_FEATURES {uninitialized_features}"""

FUNCTION_DISPATCHER = r"""{ippapi}
{ippfun}
{{
    Ipp64u _features = {package_type}GetEnabledCpuFeatures();
{dispatching_scheme}{init_check}{return_statement}
}}"""

INIT_CHECK_FORMAT = r"""
    if (_features == UNINITIALIZED_FEATURES)
    {{
        (void){package_type}Init();
        _features = {package_type}GetEnabledCpuFeatures();
        if (_features != UNINITIALIZED_FEATURES)
        {{
            {return_keyword}{function}({args});
        }}
    }}"""

DISPATCHING_SCHEME_FORMAT = r"""
    if ({cpuid} == (_features & {cpuid}))
    {{
        {return_keyword}{function}({args});
    }}"""

RETURN_VALUES = {
    "IppStatus": "ippStsCpuNotSupportedErr",
    "IppiRect": "(IppiRect) { IPP_MIN_32S / 2, IPP_MIN_32S / 2, IPP_MAX_32S, IPP_MAX_32S }",
    "void": "",
    "default": "NULL",
}

BUILD_SCRIPT = {
    WINDOWS: r""":: Generates {threading} dynamic library for {architecture} architecture
@echo off
set "OUTPUT_PATH={output_path}"
if not exist %OUTPUT_PATH% mkdir %OUTPUT_PATH%
if exist "{custom_library}.dll" del "{custom_library}.dll"

setlocal
{env_commands}
cd /d %OUTPUT_PATH%
{compile_command}
{link_command}
endlocal

if %ERRORLEVEL%==0 (
    echo Build completed!
    del /s /q /f *.obj > nul
    exit /b 0
) else (
    echo Build failed!
    exit /b 1
)
""",
    LINUX: r"""#!/bin/bash
# Generates {threading} dynamic library for {architecture} architecture
OUTPUT_PATH="{output_path}"
mkdir -m 744 -p $OUTPUT_PATH
cd $OUTPUT_PATH

rm -rf "{custom_library}.so"

{env_commands}
{compile_command}
{link_command}
if [ $? == 0 ]; then
    echo Build completed!
    rm -rf *.o
    exit 0
else
    echo Build failed!
    exit 1
fi
""",
}



"""
Copyright (C) 2018 Intel Corporation

SPDX-License-Identifier: MIT
"""

import os
from subprocess import call  # nosec

from tool import utils
from tool.generators import (
    EXPORT_GENERATORS,
    build_script_generator,
    custom_dispatcher_generator,
    main_file_generator,
    rename_header_generator,
)


def generate_script(dispatcher_type: utils.DispatcherType = utils.DispatcherType.DYNAMIC):
    """
    Generates build script
    """

    host = utils.HOST_SYSTEM
    configs = utils.CONFIGS

    package = configs[utils.PACKAGE]
    functions_list = configs[utils.FUNCTIONS_LIST]
    output_path = configs[utils.OUTPUT_PATH]

    if not os.path.exists(output_path):
        os.makedirs(output_path, 0o744)

    with open(os.path.join(output_path, utils.MAIN_FILE_NAME), "w") as main_file:
        main_file.write(main_file_generator())

    if configs[utils.CUSTOM_CPU_SET]:
        disp_folder = os.path.join(output_path, "custom_dispatcher", configs[utils.ARCHITECTURE])
        if not os.path.exists(disp_folder):
            os.makedirs(disp_folder, 0o744)

        functions_with_custom_disp = []
        for function in functions_list:
            if function not in package.functions_without_dispatcher:
                disp_file_path = os.path.join(disp_folder, f"{function}.c")
                functions_with_custom_disp.append(function)

                with open(disp_file_path, "w") as disp_file:
                    disp_file.write(custom_dispatcher_generator(function, dispatcher_type))

        if configs[utils.PREFIX]:
            with open(os.path.join(disp_folder, utils.RENAME_HEADER_NAME), "w") as rename_header:
                rename_header.write(rename_header_generator(functions_with_custom_disp))

            functions_list = [
                configs[utils.PREFIX] + func if func in functions_with_custom_disp else func for func in functions_list
            ]

    with open(os.path.join(output_path, utils.EXPORT_FILE[host]), "w") as export_file:
        EXPORT_GENERATORS[host](export_file, functions_list)

    script_path = os.path.join(output_path, configs[utils.BUILD_SCRIPT_NAME])
    with open(script_path, "w") as build_script:
        build_script.write(build_script_generator())
    os.chmod(script_path, 0o744)

    return os.path.exists(script_path)


def build():
    """
    Builds dynamic library

    :return: True if build was successful and False in the opposite case
    """
    success = generate_script()
    if not success:
        return False

    error = call([os.path.join(utils.CONFIGS[utils.OUTPUT_PATH], utils.CONFIGS[utils.BUILD_SCRIPT_NAME])])
    return False if error else True



#=========================================================================
# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License,  Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law  or agreed  to  in  writing,  software
# distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the  specific  language  governing  permissions  and
# limitations under the License.
#=========================================================================

#
# Generates jmp_ files necessary for the dispatcher
#

import re
import sys
import os
import hashlib
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--header', action='store', required=True, help='Crypto Multi-buffer Library dispatcher will be generated for functions in Header')
parser.add_argument('-o', '--out-directory', action='store', required=True, help='Output folder for generated files')
parser.add_argument('-l', '--cpu-list', action='store', required=True, help='Actual CPU list: semicolon separated string')
parser.add_argument('-c', '--compiler', action='store', required=True, help='Compiler')
args = parser.parse_args()
Header = args.header
OutDir = args.out_directory
cpulist = args.cpu_list.split(';')
compiler = args.compiler

headerID= False      ## Header ID define to avoid multiple include like: #if !defined( __IPPCP_H__ )

from gen_disp_common_crypto_mb import readNextFunction

HDR= open( Header, 'r' )
h= HDR.readlines()
HDR.close()


## keep filename only
(incdir, Header)= os.path.split(Header)

## original header name to declare external functions as internal for dispatcher
OrgH= Header

isFunctionFound = True
curLine = 0
FunType = ""
FunName = ""
FunArg = ""

while (isFunctionFound == True):

    result = readNextFunction(h, curLine, headerID)

    curLine          = result['curLine']
    FunType          = result['FunType']
    FunName          = result['FunName']
    FunArg           = result['FunArg']
    FunArgCall       = result['FunArgCall']
    isFunctionFound  = result['success']

    if (isFunctionFound == True):

        ##################################################
        ## create dispatcher C file
        ##################################################
        DISP= open( os.sep.join([OutDir, "jmp_"+FunName+"_" + hashlib.sha512(FunName.encode('utf-8')).hexdigest()[:8] + ".c"]), 'w' )

        DISP.write(f"#include <crypto_mb/{OrgH}>\n\n")
        DISP.write(f"#include <internal/common/ifma_defs.h>\n\n")

        DISP.write(f"typedef {FunType} (*MBX_FUNC_PTR){FunArg};\n\n")
        DISP.write(f"static int {FunName}_index = -1;\n")
        DISP.write(f"static int *p_{FunName}_index = &{FunName}_index;\n\n")

        DISP.write(f"extern int* _mbx_own_get_index();\n")

        DISP.write(f"extern {FunType} ini_{FunName}{FunArg};\n")

        for cpu in cpulist:
            DISP.write(f"extern {FunType} {cpu}_{FunName}{FunArg};\n")

        DISP.write(f"static MBX_FUNC_PTR arraddr[] =\n{{\n	ini_{FunName}")

        for cpu in cpulist:
            DISP.write(f",\n    {cpu}_{FunName}")

        DISP.write(f"\n}};")

        DISP.write(f"""

DLL_PUBLIC
{FunType} {FunName}{FunArg}
{{
    return (arraddr[*p_{FunName}_index + 1]){FunArgCall};
}}

{FunType} ini_{FunName}{FunArg}
{{
    p_{FunName}_index = _mbx_own_get_index();
    return (arraddr[*p_{FunName}_index + 1]{FunArgCall});
}}
""")

        DISP.close()



#=========================================================================
# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License,  Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law  or agreed  to  in  writing,  software
# distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the  specific  language  governing  permissions  and
# limitations under the License.
#=========================================================================

#
# Generates spc headers for the dispatcher. func -> ${opt}_func
#

import sys
import os
import datetime
import ntpath
import re

def path_leaf(path):
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)

sys.path.append(os.path.join(sys.path[0], '../dispatcher'))

from gen_disp_common_crypto_mb import readNextFunction

Header  = sys.argv[1]
OutDir  = sys.argv[2]

Header = os.path.abspath(Header)
Filename = ""

HDR= open(Header, 'r')
h= HDR.readlines()
HDR.close()

headerID= False
FunName = ""


if not os.path.exists(OutDir):
  os.makedirs(OutDir)


Filename = re.sub(r'.h','',path_leaf(Header))

OutFile  = os.sep.join([OutDir, Filename + "_cpuspc" + ".h"])

year = datetime.datetime.today()

OUT= open( OutFile, 'w' )
OUT.write(f"""/*******************************************************************************
* Copyright {year} Intel Corporation
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*******************************************************************************/

""")

curLine = 0
isFunctionFound = True

while (isFunctionFound):

	result = readNextFunction(h, curLine, headerID)

	curLine         = result['curLine']
	FunName         = result['FunName']
	isFunctionFound = result['success']

	if (isFunctionFound):
		OUT.write(f"#define {FunName} OWNAPI({FunName})\n")
OUT.close()



#=========================================================================
# Copyright (C) 2017 Intel Corporation
#
# Licensed under the Apache License,  Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law  or agreed  to  in  writing,  software
# distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the  specific  language  governing  permissions  and
# limitations under the License.
#=========================================================================

#
# Intel(R) Cryptography Primitives Library
#

import re
import sys
import os
import hashlib
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--header', action='store', required=True, help='Intel Cryptography Primitives Library dispatcher will be generated for functions in Header')
parser.add_argument('-o', '--out-directory', action='store', required=True, help='Output folder for generated files')
parser.add_argument('-l', '--cpu-list', action='store', required=True, help='Actual CPU list: semicolon separated string')
parser.add_argument('-c', '--compiler', action='store', required=True, help='Compiler')
args = parser.parse_args()
Header = args.header
OutDir = args.out_directory
cpulist = args.cpu_list.split(';')
compiler = args.compiler

headerID= False      ## Header ID define to avoid multiple include like: #if !defined( __IPPCP_H__ )

from gen_disp_common import readNextFunction

HDR= open( Header, 'r' )
h= HDR.readlines()
HDR.close()


## keep filename only
(incdir, Header)= os.path.split(Header)

## original header name to declare external functions as internal for dispatcher
OrgH= Header

isFunctionFound = True
curLine = 0
FunType = ""
FunName = ""
FunArg = ""

if(compiler == "GNU" or compiler == "Clang"  or compiler == "IntelLLVM"):

    while (isFunctionFound == True):

        result = readNextFunction(h, curLine, headerID)

        curLine         = result['curLine']
        FunType         = result['FunType']
        FunName         = result['FunName']
        FunArg          = result['FunArg']
        isFunctionFound = result['success']

        if (isFunctionFound == True):

            ##################################################
            ## create dispatcher files ASM
            ##################################################
            ASMDISP= open( os.sep.join([OutDir, "jmp_" + FunName+"_" + hashlib.sha512(FunName.encode('utf-8')).hexdigest()[:8] +".asm"]), 'w' )

            ASMDISP.write('%include "noexec_stack.inc"\n\n');

            # Symbol type setting for extern functions initially appeared in version 2.15
            ASMDISP.write("%if ((__NASM_MAJOR__ > 2) || ((__NASM_MAJOR__ == 2) && (__NASM_MINOR__ > 14)))\n");
            ASMDISP.write("  %xdefine elf_symbol_type :function\n");
            ASMDISP.write("%else\n");
            ASMDISP.write("  %xdefine elf_symbol_type\n");
            ASMDISP.write("%endif\n");
            for cpu in cpulist:
                ASMDISP.write("extern "+cpu+"_"+FunName+"%+elf_symbol_type\n")

            ASMDISP.write("extern ippcpJumpIndexForMergedLibs\n")
            ASMDISP.write("extern ippcpSafeInit%+elf_symbol_type\n\n")
            ASMDISP.write("""
segment .data
align 8
dq  .Lin_{FunName}
.Larraddr_{FunName}:
""".format(FunName=FunName))

            for cpu in cpulist:
                ASMDISP.write("    dq "+cpu+"_"+FunName+"\n")

            ASMDISP.write("""
segment .text
global {FunName}:function ({FunName}.LEnd{FunName} - {FunName})
.Lin_{FunName}:
    {endbr64}
    call ippcpSafeInit wrt ..plt
    align 16

{FunName}:
    {endbr64}
    mov     rax, qword [rel ippcpJumpIndexForMergedLibs wrt ..gotpc]
    movsxd  rax, dword [rax]
    lea     r11, [rel .Larraddr_{FunName}]
    mov     r11, qword [r11+rax*8]
    jmp     r11
.LEnd{FunName}:
""".format(FunName=FunName, endbr64='db 0xf3, 0x0f, 0x1e, 0xfa'))
            ASMDISP.close()

else:
    while (isFunctionFound == True):

        result = readNextFunction(h, curLine, headerID)

        curLine         = result['curLine']
        FunType         = result['FunType']
        FunName         = result['FunName']
        FunArg          = result['FunArg']
        isFunctionFound = result['success']

        if (isFunctionFound == True):

            ##################################################
            ## create dispatcher files: C file with inline asm
            ##################################################
            DISP= open( os.sep.join([OutDir, "jmp_"+FunName+"_" + hashlib.sha512(FunName.encode('utf-8')).hexdigest()[:8] + ".c"]), 'w' )

            DISP.write("""#include "ippcp.h"\n\n#pragma warning(disable : 1478 1786) // deprecated\n\n""")

            DISP.write("typedef "+FunType+" (*IPP_PROC)"+FunArg+";\n\n")
            DISP.write("extern int ippcpJumpIndexForMergedLibs;\n")
            DISP.write("extern IppStatus ippcpSafeInit( void );\n\n")

            DISP.write("extern "+FunType+" in_"+FunName+FunArg+";\n")

            for cpu in cpulist:
                DISP.write("extern "+FunType+" "+cpu+"_"+FunName+FunArg+";\n")

            DISP.write("static IPP_PROC arraddr[] =\n{{\n	(IPP_PROC)in_{}".format(FunName))

            for cpu in cpulist:
                DISP.write(",\n	(IPP_PROC)"+cpu+"_"+FunName+"")

            DISP.write("\n};")

            DISP.write("""
#undef  IPPAPI
#define IPPAPI(type,name,arg) __declspec(naked) type name arg
IPPAPI({FunType}, {FunName},{FunArg})
{{
    register unsigned long long i __asm__("rax");
    i = (unsigned long long)arraddr[ippcpJumpIndexForMergedLibs+1];
    _asm{{ jmp rax }}
}}
IPPAPI({FunType}, in_{FunName},{FunArg})
{{
   __asm{{
        call ippcpSafeInit
        mov  rax, qword ptr [{FunName}]
        jmp  rax
  }}
}}
""".format(FunType=FunType, FunName=FunName, FunArg=FunArg))

            DISP.close()



#=========================================================================
# Copyright (C) 2017 Intel Corporation
#
# Licensed under the Apache License,  Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law  or agreed  to  in  writing,  software
# distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the  specific  language  governing  permissions  and
# limitations under the License.
#=========================================================================

#
# Intel(R) Cryptography Primitives Library
#

import re
import sys
import os
import hashlib
import argparse

parser = argparse.ArgumentParser()
parser.add_argument('-i', '--header', action='store', required=True, help='Intel Cryptography Primitives Library dispatcher will be generated for functions in Header')
parser.add_argument('-o', '--out-directory', action='store', required=True, help='Output folder for generated files')
parser.add_argument('-l', '--cpu-list', action='store', required=True, help='Actual CPU list: semicolon separated string')
parser.add_argument('-c', '--compiler', action='store', required=True, help='Compiler')
args = parser.parse_args()
Header = args.header
OutDir = args.out_directory
cpulist = args.cpu_list.split(';')
compiler = args.compiler

headerID= False      ## Header ID define to avoid multiple include like: #if !defined( __IPPCP_H__ )

from gen_disp_common import readNextFunction

HDR= open( Header, 'r' )
h= HDR.readlines()
HDR.close()


## keep filename only
(incdir, Header)= os.path.split(Header)

## original header name to declare external functions as internal for dispatcher
OrgH= Header

isFunctionFound = True
curLine = 0
FunName = ""
FunArg = ""

if(compiler == "GNU" or compiler == "Clang" or compiler == "IntelLLVM"):
      while (isFunctionFound == True):

            result = readNextFunction(h, curLine, headerID)

            curLine         = result['curLine']
            FunName         = result['FunName']
            FunArg          = result['FunArg']
            isFunctionFound = result['success']

            if (isFunctionFound == True):

                  ##################################################
                  ## create dispatcher files ASM
                  ##################################################
                  ASMDISP= open( os.sep.join([OutDir, "jmp_" + FunName+"_" + hashlib.sha512(FunName.encode('utf-8')).hexdigest()[:8] +".asm"]), 'w' )

                  ASMDISP.write('%include "noexec_stack.inc"\n\n');

                  # Symbol type setting for extern functions initially appeared in version 2.15
                  ASMDISP.write("%if ((__NASM_MAJOR__ > 2) || ((__NASM_MAJOR__ == 2) && (__NASM_MINOR__ > 14)))\n");
                  ASMDISP.write("  %xdefine elf_symbol_type :function\n");
                  ASMDISP.write("%else\n");
                  ASMDISP.write("  %xdefine elf_symbol_type\n");
                  ASMDISP.write("%endif\n");
                  for cpu in cpulist:
                      ASMDISP.write("extern "+cpu+"_"+FunName+"%+elf_symbol_type\n")

                  ASMDISP.write("extern ippcpJumpIndexForMergedLibs\n")
                  ASMDISP.write("extern ippcpInit%+elf_symbol_type\n\n")

                  ASMDISP.write("""
segment .data
align 4
dd  .Lin_{FunName}
.Larraddr_{FunName}:
""".format(FunName=FunName))

                  for cpu in cpulist:
                        ASMDISP.write("    dd "+cpu+"_"+FunName+"\n")

                  ASMDISP.write("""

segment .text
extern _GLOBAL_OFFSET_TABLE_
global {FunName}:function ({FunName}.LEnd{FunName} - {FunName})
.Lin_{FunName}:
    {endbr32}
    push ebx
    mov ebx, eax
    call  ippcpInit wrt ..plt
    pop ebx
    align 16
{FunName}:
    {endbr32}
    call .L1
.L1:
      pop   eax
      add   eax, _GLOBAL_OFFSET_TABLE_ + $$ - .L1 wrt ..gotpc
      mov   edx, [eax + ippcpJumpIndexForMergedLibs wrt ..got]
      mov   edx, [edx]
      jmp   dword [eax + edx*4 + .Larraddr_{FunName} wrt ..gotoff]
.LEnd{FunName}:
""".format(FunName=FunName, endbr32='db 0xf3, 0x0f, 0x1e, 0xfb'))
            ASMDISP.close()
else:

      while (isFunctionFound == True):

            result = readNextFunction(h, curLine, headerID)

            curLine         = result['curLine']
            FunName         = result['FunName']
            FunArg          = result['FunArg']
            isFunctionFound = result['success']

            if (isFunctionFound == True):

                  ##################################################
                  ## create dispatcher files: C file with inline asm
                  ##################################################
                  DISP= open( os.sep.join([OutDir, "jmp_"+FunName+"_" + hashlib.sha512(FunName.encode('utf-8')).hexdigest()[:8] + ".c"]), 'w' )

                  DISP.write("""#include "ippcpdefs.h"\n\n""")

                  DISP.write("typedef void (*IPP_PROC)(void);\n\n")
                  DISP.write("extern int ippcpJumpIndexForMergedLibs;\n")
                  DISP.write("extern IPP_CALL ippcpInit();\n\n")

                  DISP.write("extern IppStatus IPP_CALL in_"+FunName+FunArg+";\n")

                  for cpu in cpulist:
                        DISP.write("extern IppStatus IPP_CALL "+cpu+"_"+FunName+FunArg+";\n")

                  DISP.write("""
__asm( "  .data");
__asm( "    .align 4");
__asm( "arraddr:");
__asm( "    .long	in_{FunName}");""".format(FunName=FunName))
                  size = 4
                  for cpu in cpulist:
                        size = size + 4
                        DISP.write("""\n__asm( "    .long	{cpu}_{FunName}");""".format(FunName=FunName, cpu=cpu))

                  DISP.write("""
__asm( "    .type	arraddr,@object");
__asm( "    .size	arraddr,{size}");
__asm( "  .data");\n""".format(size=size))

                  DISP.write("""
#undef  IPPAPI
#define IPPAPI(type,name,arg) __declspec(naked) void IPP_CALL name arg
__declspec(naked) IPP_PROC {FunName}{FunArg}
{{
    __asm( ".L0: call .L1");
    __asm( ".L1: pop %eax");
    __asm( "add $_GLOBAL_OFFSET_TABLE_ - .L1, %eax" );
    __asm( "movd ippcpJumpIndexForMergedLibs@GOT(%eax), %xmm0" );
    __asm( "movd %xmm0, %edx" );
    __asm( "mov (%edx), %edx" );
    __asm( "mov (arraddr@GOTOFF+4)(%eax,%edx,4), %edx" );
    __asm( "jmp *%edx" );
    __asm( ".global in_{FunName}" );
    __asm( "in_{FunName}:" );
    {endbr32}
    __asm( "push %ebx" );
    __asm( "mov %eax, %ebx" );
    __asm( "call ippcpInit@PLT" );
    __asm( "pop %ebx" );
    __asm( "jmp .L0" );
}}
""".format(FunName=FunName, FunArg=FunArg, endbr32='__asm( ".byte 0xf3, 0x0f, 0x1e, 0xfb" );'))

      DISP.close()




#=========================================================================
# Copyright (C) 2024 Intel Corporation
#
# Licensed under the Apache License,  Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# 	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law  or agreed  to  in  writing,  software
# distributed under  the License  is  distributed  on  an  "AS IS"  BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the  specific  language  governing  permissions  and
# limitations under the License.
#=========================================================================

#
# Generates headers for 1cpu library. func -> ${opt}_func
#

import sys
import os
import datetime
import ntpath
import re

def path_leaf(path):
    head, tail = ntpath.split(path)
    return tail or ntpath.basename(head)

sys.path.append(os.path.join(sys.path[0], '../dispatcher'))

from gen_disp_common_crypto_mb import readNextFunction

Header  = sys.argv[1]
OutDir  = sys.argv[2]

Header = os.path.abspath(Header)
Filename = ""

HDR= open(Header, 'r')
h= HDR.readlines()
HDR.close()

headerID= False
FunName = ""

Filename = re.sub(r'.h','',path_leaf(Header))

if not os.path.exists(OutDir):
  os.makedirs(OutDir)

Filenames=["k1", "l9"]

year = datetime.datetime.today()

for name in Filenames:
  OutFile  = os.sep.join([OutDir, Filename + "_"+ name + ".h"])
  
  OUT= open( OutFile, 'w' )
  OUT.write(f"""/*******************************************************************************
  * Copyright {year} Intel Corporation
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
  * You may obtain a copy of the License at
  *
  *     http://www.apache.org/licenses/LICENSE-2.0
  *
  * Unless required by applicable law or agreed to in writing, software
  * distributed under the License is distributed on an "AS IS" BASIS,
  * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  * See the License for the specific language governing permissions and
  * limitations under the License.
  *******************************************************************************/

  """)

  curLine = 0
  isFunctionFound = True

  while (isFunctionFound):

    result = readNextFunction(h, curLine, headerID)

    curLine         = result['curLine']
    FunName         = result['FunName']
    isFunctionFound = result['success']

    if (isFunctionFound):
      OUT.write(f"#define {FunName} {name}_{FunName}\n")
  OUT.close()
